//-------------------------------------------------------------------
// OWL Extensions (OWLEXT) Class Library
// Copyright(c) 1996 by Manic Software.
// All rights reserved.
//
//$Revision: 1.1.2.1 $
//$Author: jogybl $
//$Date: 2007/08/23 08:17:54 $
//
// Provides the classes TGscThread, which is a progress-notifying thread class,
// and TThreadProgressDlg, which is a Progress dialog that handles thread
// creation and destruction.
//
// Works in both 16- and 32- bit environments
//
// Original code:
// Copyright (c) 1997 Rich Goldstein, MD
// goldstei@interport.net
// May be used without permission, if appropriate credit is given
//-------------------------------------------------------------------
#include <owlext\pch.h>
#pragma hdrstop

#include <owlext/ThredPrg.h>


OWLEXT_BEGIN_NAMESPACE


TProgressThread::TProgressThread(UINT id, HWND hTarget) :
#if !defined(BI_PLAT_WIN16)
   TThread(),
#endif
     _hTarget(hTarget)
   , _id(id)
#if defined(BI_PLAT_WIN16)
   , _stat(Created)
   , _terminationRequested(false)
   , _exitCode(0)
#endif
{ }



void TProgressThread::NotifyStart(uint32 len)
{
   if (_hTarget)
      ::PostMessage(_hTarget, WM_GSC_PROGRESS_START, (WPARAM)_id, (LPARAM)len);
#if defined(BI_PLAT_WIN16)
   GetApplicationObject()->PumpWaitingMessages();
#endif
}

void TProgressThread::NotifyProgress(uint32 curr)
{
   if (_hTarget)
      ::PostMessage(_hTarget, WM_GSC_PROGRESS, (WPARAM)_id, (LPARAM)curr);
#if defined(BI_PLAT_WIN16)
   GetApplicationObject()->PumpWaitingMessages();
#endif
}

void TProgressThread::NotifyEnd()
{
   if (_hTarget)
      ::PostMessage(_hTarget, WM_GSC_PROGRESS_END, (WPARAM)_id, 0);
#if defined(BI_PLAT_WIN16)
   GetApplicationObject()->PumpWaitingMessages();
#endif
}

#if defined(BI_PLAT_WIN16)
bool TProgressThread::ShouldTerminate() const
{
   GetApplicationObject()->PumpWaitingMessages();
   return _terminationRequested;
}

// Sleep has a different meaning in the 16 bit version!
bool TProgressThread::Sleep(long timeMS, bool)
{
   if (!timeMS)
      return true;
   DWORD timeStart = ::GetTickCount();
   DWORD timeStop = timeStart;
   while (timeStop - timeStart < timeMS)
   {
      GetApplicationObject()->PumpWaitingMessages();
      timeStop = ::GetTickCount();
   }
   return true;
}

uint32 TProgressThread::Start()
{
   _stat = Running;
   _exitCode = Run();
   _stat = Finished;
   return _exitCode;
}
#endif   // BI_PLAT_WIN16


//
DEFINE_RESPONSE_TABLE1(TThreadProgressDlg, TProgressDlg)
   EV_COMMAND(IDCANCEL, CmCancel),
#if defined(BI_PLAT_WIN16)
   EV_MESSAGE(WM_GSC_PROGRESS_BEGINTHREAD, EvProgressBeginThread),
#endif
END_RESPONSE_TABLE;

TThreadProgressDlg::TThreadProgressDlg(TWindow* parent, const char* text,
                     TResId id, TModule* mod)
   : TProgressDlg(parent, text, id, mod)
   , _othread(0)
#if defined(BI_PLAT_WIN16)
   , _cancelled(false)
#endif
{}

TThreadProgressDlg::~TThreadProgressDlg()
{
   delete _othread;
}

void TThreadProgressDlg::SetupWindow()
{
   TProgressDlg::SetupWindow();
   _othread = BuildThread();
#if !defined(BI_PLAT_WIN16)
   if (_othread)
      _othread->Start();
   else
      CloseWindow(IDCANCEL);
#else
   ::PostMessage(GetHandle(), WM_GSC_PROGRESS_BEGINTHREAD, 0, 0);
#endif
}

void TThreadProgressDlg::CleanupWindow()
{
#if !defined(BI_PLAT_WIN16)
   if (_othread && _othread->GetStatus() != TThread::Finished)
      _othread->TerminateAndWait(TThread::NoLimit);
#endif
   TProgressDlg::CleanupWindow();
}

bool TThreadProgressDlg::CanClose()
{
   if (_othread && _othread->GetStatus() != TProgressThread::Finished)
      return false;
   else
      return true;
}

void TThreadProgressDlg::CmCancel()
{
   #if defined(BI_PLAT_WIN16)
   if (_othread && _othread->GetStatus() != TProgressThread::Finished)
   {
      _cancelled = true;
      _othread->Terminate();
   }
   #else
   if (_othread && _othread->GetStatus() != TThread::Finished)
      _othread->TerminateAndWait(TThread::NoLimit);
   TProgressDlg::CmCancel();
   #endif
}

#if defined (BI_PLAT_WIN16)
LRESULT TThreadProgressDlg::EvProgressBeginThread(WPARAM, LPARAM)
{
   if (_othread)
      _othread->Start();
   else
      _cancelled = true;
   CloseWindow(_cancelled ? IDCANCEL : IDOK);
}
#endif


OWLEXT_END_NAMESPACE



//
// Test code follows
#if defined(TEST_THE_CODE)
#include <owl/statusba.h>
#include <owl/controlg.h>
#include "gsc/testprg.rh"
#include "gsc/gaugegad.h"

OWLEXT_USING_NAMESPACE

class SillyThread : public TProgressThread
{
   public:
      SillyThread(HWND hWnd, int count)
         : TProgressThread(0, hWnd), _count(count)
         {}

      int Run()
      {
         NotifyStart(_count);
         for (int i = 0; i < _count && !ShouldTerminate(); ++i)
         {
            Sleep(100);
            NotifyProgress(i);
         }
         NotifyEnd();
         return 0;
      }

   protected:
      int   _count;
};

class SillyProgressDlg : public TThreadProgressDlg
{
   public:
      SillyProgressDlg(TWindow* parent)
         : TThreadProgressDlg(parent, "Running Silly Loop")
      {}

      TProgressThread* BuildThread() { return new SillyThread(GetHandle(), 50); }
};



class TProgressFrame : public TDecoratedFrame
{
   public:
      TProgressFrame(TWindow* parent,
                const char far* title,
                TWindow* clientWnd,
                bool trackMenuSelection = false,
                TModule* module = 0);

   protected:
      void CmBeginThread();
      void CmEndThread();
      void CmDialogThread();

      void CleanupWindow();

      LRESULT EvProgressStart(WPARAM, LPARAM range)
         { _gauge->SetRange(0, (int)range); return 0; }
      LRESULT EvProgress(WPARAM, LPARAM value)
         { _gauge->SetValue((int)value); return 0; }
      LRESULT EvProgressEnd(WPARAM, LPARAM);

      TStatusBar* _status;
      TGaugeGadget*   _gauge;
      // TGauge*     _gauge;

      TProgressThread* _othread;

   DECLARE_RESPONSE_TABLE(TProgressFrame);
};

DEFINE_RESPONSE_TABLE1(TProgressFrame, TDecoratedFrame)
   EV_MESSAGE(WM_GSC_PROGRESS_START, EvProgressStart),
   EV_MESSAGE(WM_GSC_PROGRESS, EvProgress),
   EV_MESSAGE(WM_GSC_PROGRESS_END, EvProgressEnd),
   EV_COMMAND(CM_BEGINTHREAD, CmBeginThread),
   EV_COMMAND(CM_ENDTHREAD, CmEndThread),
   EV_COMMAND(CM_DIALOGTHREAD, CmDialogThread),
END_RESPONSE_TABLE;

TProgressFrame::TProgressFrame(TWindow* parent,
                const char far* title,
                TWindow* clientWnd,
                bool trackMenuSelection,
                TModule* module)
   : TDecoratedFrame(parent, title, clientWnd, trackMenuSelection, module)
   , _othread(0)
{
            _status = new TStatusBar(0, TGadget::Recessed,
                                TStatusBar::CapsLock |
                                TStatusBar::NumLock |
                                TStatusBar::Overtype);
            // Create a gauge control
            _gauge = new TGaugeGadget(1, "%d%%", 96);
            _status->Insert(*_gauge);
            Insert(*_status, TDecoratedFrame::Bottom);
}

void TProgressFrame::CleanupWindow()
{
   if (_othread)
   {
      if (_othread->GetStatus() != TProgressThread::Finished)
         _othread->TerminateAndWait(TProgressThread::NoLimit);
      delete _othread;
   }
   TDecoratedFrame::CleanupWindow();
}


LRESULT TProgressFrame::EvProgressEnd(WPARAM, LPARAM)
{
   int l, h;
   _gauge->GetRange(l, h);
   _gauge->SetValue(l);

   // Wait for thread to end, then delete the thread object
   if (_othread)
   {
      if (_othread->GetStatus() != TProgressThread::Finished)
         _othread->TerminateAndWait(TProgressThread::NoLimit);
      delete _othread;
      // Prevent re-entrancy in 16 bit versions, and multiple
      // instances of the thread in 32 bit versions
      _othread = 0;
   }

   return 0;
}

void TProgressFrame::CmDialogThread()
{
   SillyProgressDlg(this).Execute();
}

void TProgressFrame::CmEndThread()
{
   if (_othread && _othread->GetStatus() != TProgressThread::Finished)
      _othread->Terminate();
}

void TProgressFrame::CmBeginThread()
{
   if (_othread)
   {
      MessageBox("A thread is already running.");
      return;
   }
   _othread = new SillyThread(GetHandle(), 100);

   // In 16 bit version, be careful to avoid re-entrancy!
   if (_othread)
      _othread->Start();
}

class App : public TApplication
{
	public:
		App()
      : TApplication("Thread Progress Demo")
		{ }

      ~App()
      { }


		void InitMainWindow()
		{
			TFrameWindow* frame = new TProgressFrame(0,
                                                GetName(),
                                                new TWindow(0, ""),
                                                false);
         frame->AssignMenu(MAINMENU);
			SetMainWindow(frame);

		}
};

int OwlMain (int , char* [])
{
	try {
		App app;
		return app.Run();
	}
	catch (xmsg& x) {
		::MessageBox(0, x.why().c_str(), "Exception", MB_OK);
	}
	return -1;
}
#endif // TEST_THE_CODE
//

